from .bgptools import update, search_asn_name, search_asn_as_ip
import roul.ip.radix

import ipaddress

ASNS: dict[int, str] = {}
TABLE_IPV4: roul.ip.radix.RadixTree = roul.ip.radix.RadixTree(bit_length=32)
TABLE_IPV6: roul.ip.radix.RadixTree = roul.ip.radix.RadixTree(bit_length=128)
UPDATED_AT: float = 0


import roul
import roul.ip
import roul.ip.radix
import roul.asn

from requests import get
import ipaddress
import json
import time
import csv


BGP_TOOLS_ASN_URL = "https://bgp.tools/asns.csv"
BGP_TOOLS_TABLE_URL = "https://bgp.tools/table.jsonl"
UA = "SPARCS@KAIST sparcs.org - noc@sparcs.net"

def update():
    asns = get(BGP_TOOLS_ASN_URL, headers={"User-Agent": UA}).text.splitlines()[1:]
    tables = get(BGP_TOOLS_TABLE_URL, headers={"User-Agent": UA}).text.splitlines()

    new_asns: dict[int, str] = {}
    new_table_ipv4 = roul.ip.radix.RadixTree(bit_length=32)
    new_table_ipv6 = roul.ip.radix.RadixTree(bit_length=128)

    reader = csv.reader(asns[1:])
    for row in reader:
        try:
            new_asns[int(row[0][2:])] = row[1].replace("\n", "")
        except Exception as e:
            print(f"Warning: Skipping invalid ASN line -> {row} ({e})")

    for row in tables:
        try:
            record = json.loads(row)
            cidr: str = record['CIDR']
            asn: int = record['ASN']

            if roul.ip.is_ipv4(cidr):
                new_table_ipv4.add(cidr, asn)
            elif roul.ip.is_ipv6(cidr):
                new_table_ipv6.add(cidr, asn)
            else:
                raise TypeError("Network is neither IPv4 nor IPv6")
            
        # except (json.JSONDecodeError, KeyError, ValueError) as e:
        #     print(f"Warning: Skipping invalid line -> {row} ({e})")
        except Exception as e:
            print(f"Warning: Skipping invalid line -> {row} ({e})")
            raise e


    roul.asn.ASNS = new_asns
    roul.asn.TABLE_IPV4 = new_table_ipv4
    roul.asn.TABLE_IPV6 = new_table_ipv6
    roul.asn.UPDATED_AT = time.time()

    del asns, tables

def search_asn_as_ip(ipaddr) -> int:
    """
    Find the ASN of the given IP address.

    Args:
        ipaddr (str): The IP address to find the ASN of.

    Return:
        int: The ASN of the given IP address.
    """
    if not roul.ip.is_valid(ipaddr):
        raise ValueError("Invalid IP address")
    
    ipaddr_ipnw = ipaddress.ip_address(ipaddr)

    if roul.ip.is_ipv4(ipaddr_ipnw):
        return roul.asn.TABLE_IPV4.search_best(ipaddr)
    elif roul.ip.is_ipv6(ipaddr_ipnw):
        return roul.asn.TABLE_IPV6.search_best(ipaddr)
    else:
        raise ValueError("IP address is neither IPv4 nor IPv6")
    
def search_asn_name(asn: int) -> str:
    """
    Search the name of the given ASN.
    """

    if asn not in roul.asn.ASNS:
        raise ValueError(f"ASN {asn} not found. It may need to update()")

    return roul.asn.ASNS[asn]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    install_requires = fh.read().splitlines()

setuptools.setup(
    name="roul",
    version="0.0.1-pre1",
    author="DoyunShin",
    author_email="op@doyun.me",
    description="ROUL Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10.0",
    install_requires=install_requires,
    # entry_points={
    #     "console_scripts": [
    #         "ptunnel = ptunnel.__main__:main",
    #     ]
    # },
)


import roul.ip

import ipaddress

class RadixNode:
    def __init__(self):
        self.right: RadixNode | None = None
        self.left: RadixNode | None = None
        self.data: int = -1

class RadixTree:
    def __init__(self, bit_length: int):
        self.root = RadixNode()
        self.bit_length = bit_length

    def add(self, cidr: str, asn: int):
        ip_int = roul.ip.to_int(roul.ip.network_to_address(cidr))
        prefixlen = roul.ip.prefixlen(cidr)
        
        node = self.root
        
        for i in range(prefixlen):
            bit = (ip_int >> (self.bit_length - 1 - i)) & 1
            
            if bit == 1:
                if not node.right:
                    node.right = RadixNode()
                node = node.right
            else:
                if not node.left:
                    node.left = RadixNode()
                node = node.left
        
        node.data = asn

    def search_best(self, ip_addr: str) -> int:
        ip_int = int(ipaddress.ip_address(ip_addr))
        
        node = self.root
        last_match = None
        
        if self.root.data:
            last_match = self.root.data

        for i in range(self.bit_length):
            bit = (ip_int >> (self.bit_length - 1 - i)) & 1
            
            if bit == 1:
                node = node.right
            else:
                node = node.left
            
            if not node:
                break
            
            if node.data:
                last_match = node.data
        
        if last_match == None:
            raise ValueError("No match found")
        
        return last_match



import ipaddress

def is_valid(ipaddr: str):
    """Check if the IP address is valid."""

    try:
        if "/" in ipaddr:
            ipaddress.ip_network(ipaddr)
        else:
            ipaddress.ip_address(ipaddr)
        return True
    except ValueError:
        return False

def is_ipv4(ipaddr: str | ipaddress.IPv4Address | ipaddress.IPv6Address | ipaddress.IPv4Network | ipaddress.IPv6Network):
    """Check if the IP address is IPv4."""

    if isinstance(ipaddr, str) and is_valid(ipaddr):
        if "/" in ipaddr:
            ipaddr = ipaddress.ip_network(ipaddr)
        else:
            ipaddr = ipaddress.ip_address(ipaddr)

    try:
        return isinstance(ipaddr, ipaddress.IPv4Address) or isinstance(ipaddr, ipaddress.IPv4Network)
    except ValueError:
        return False

def is_ipv6(ipaddr: str | ipaddress.IPv4Address | ipaddress.IPv6Address | ipaddress.IPv4Network | ipaddress.IPv6Network):
    """Check if the IP address is IPv6."""

    if isinstance(ipaddr, str) and is_valid(ipaddr):
        if "/" in ipaddr:
            ipaddr = ipaddress.ip_network(ipaddr)
        else:
            ipaddr = ipaddress.ip_address(ipaddr)

    try:
        return isinstance(ipaddr, ipaddress.IPv6Address) or isinstance(ipaddr, ipaddress.IPv6Network)
    except ValueError:
        return False

def to_int(ipaddr: str | ipaddress.IPv4Address | ipaddress.IPv6Address):
    return int(ipaddress.ip_address(ipaddr))

def to_str(ipaddr: int | ipaddress.IPv4Address | ipaddress.IPv6Address):
    return str(ipaddress.ip_address(ipaddr))

def is_in_network(ipaddr: str | int, network: str | int):
    """
    Check if the IP address is in the given network.

    Args:
        ipaddr (str | int): The IP address to check. (e.g. "1.1.1.1"))
        network (str | int): The network to check against. (e.g. "1.1.1.0/24")

    Returns:
        bool: True if the IP address is in the network, False otherwise.
    """
    return ipaddress.ip_address(ipaddr) in ipaddress.ip_network(network)

def prefixlen(cidr: str | ipaddress.IPv4Network | ipaddress.IPv6Network):
    """
    Get the prefix length of the given IP address.

    Args:
        cidr: The IP address to get the prefix length of.

    Returns:
        int: The prefix length of the given IP address.
    """
    return ipaddress.ip_network(cidr).prefixlen

def network_to_address(cidr: str | ipaddress.IPv4Network | ipaddress.IPv6Network):
    return ipaddress.ip_network(cidr).network_address
